import numpy
from . import interpolate
from . import geometry

def arc_length(tck, num_points=None):
    """Approximate the arc-length of spline (t,c,k) by evaluating it at num_points
    positions and calculating the length of the resulting polyline.
    If num_points is None, try to guess a sane default."""
    points = _get_points(tck, num_points)
    distances = geometry.cumulative_distances(points, unit=False)
    return distances[-1]


def perpendiculars(tck, num_points=None, unit=True):
    """Return vectors perpendicular to a 2D parametric spline.

    Parameters:
    tck: parametric spline tuple
    num_points: number of equally-spaced points to evaluate perpendiculars at,
        or None, which causes the code to try to guess a good number of points.
    unit: normalize prependiculars to unit length.

    Returns: array of shape (num_points, 2), containing num_points different
        2D vectors describing each perpendicular.
    """
    der = _get_points(tck, num_points, derivative=1) # shape = (length, 2)
    perpendiculars = numpy.empty_like(der)
    perpendiculars[:,0] = -der[:,1]
    perpendiculars[:,1] = der[:,0]
    if unit:
        perpendiculars /= numpy.sqrt((perpendiculars**2).sum(axis=1))[:,numpy.newaxis]
    return perpendiculars


def outline(tck, radius_tck, num_points=None):
    """Given a shape defined by a centerline spline and a radial profile spline,
    return a polygonal outline defined by sampling the centerline num_points
    along its profile and displacing it left and right of the centerline by an
    amount defined by the radius_tck.

    If num_points is None, try to guess a sane default.

    Returns: left, right, outline, where left and right are the displaced points
    (shape=(num_points, 2)) and outline is the full polygon (shape=(2*num_points, 2))
    """
    points, radii = _get_points_and_radii(tck, radius_tck, num_points)
    perps = perpendiculars(tck, num_points=len(points))
    offsets = perps * radii[:,numpy.newaxis]
    left = points + offsets
    right = points - offsets
    outline = numpy.concatenate([left, right[::-1]], axis=0)
    return left, right, outline


def area(tck, radius_tck, num_points):
    """Given a shape defined by a centerline spline and a radial profile spline,
    estimate its area by converting to a polygon at num_points along the centerline
    and applying the standard polygon area formula.

    If num_points is None, try to guess a sane default.
    """
    polygon = outline(tck, radius_tck, num_points)
    xs = points[:,0]
    ys = points[:,1]
    y_forward = numpy.roll(ys, -1, axis = 0)
    y_backward = numpy.roll(ys, 1, axis = 0)
    return numpy.absolute(numpy.sum(xs * (y_backward - y_forward)) / 2.0)


def volume_and_surface_area(tck, radius_tck, num_points):
    """Given a shape defined by a centerline spline and a radial profile spline,
    estimate the volume and surface area of a 3D form constructed by revolution
    around the centerline, by sampling the positions and radii at num_points
    along the centerline, assuming that the shape of revolution is made of
    and conical frustrums defined by these radii, and applying the standard
    formulae for their surface area and volume.

    If num_points is None, try to guess a sane default.

    Returns: volume, surface_area"""
    points, radii = _get_points_and_radii(tck, radius_tck, num_points)
    lengths = numpy.sqrt(((points[:-1] - points[1:])**2).sum(axis=1))
    # formulae from http://en.wikipedia.org/wiki/Frustum
    r1 = radii[:-1]
    r2 = radii[1:]
    r12 = r1**2
    r22 = r2**2
    h = lengths
    v = numpy.pi/3 * (h * (r12 + r22 + r1*r2)).sum()
    a = numpy.pi * (numpy.sqrt((r12-r22)**2 + (h*(r1+r2))**2).sum() + r12[0] + r22[-1])
    return v, a


def length_and_max_width(tck, radius_tck, num_points):
    """Given a shape defined by a centerline spline and a radial profile spline,
    estimate its length and maximum width by sampling at num_points along the
    spline.

    If num_points is None, try to guess a sane default.

    Returns: length, max_width
    """
    points, radii = _get_points_and_radii(tck, radius_tck, num_points)
    return numpy.sqrt(((spine[:-1] - spine[1:])**2).sum(axis=1)).sum(), widths.max()


def _get_points(tck, num_points=None, derivative=0):
    """Evaluate a spline (or its derivative) at a given number of points.

    If the number of points is not specified, make a sane guess such that the
    resulting curve will be approximately as smooth as the spline.
    Specifically, use the maximum parameter value of the spline or 100,
    whichever is greater. The maximum parameter value makes sense only if the
    curve is parameterized with something close to the 'natural parameter',
    which interpolate.fit_spline() tries to guarantee."""
    if num_points is None:
        num_points = max(100, int(round(tck[0].max())))
    points = interpolate.spline_interpolate(tck, num_points, derivative)
    return points


def _get_points_and_radii(tck, radius_tck, num_points):
    if tck[1].ndim != 2 and tck[1].shape[1] != 2:
        raise ValueError('tck must be a two-dimensional parametric spline')
    if radius_tck[1].ndim != 1:
        raise ValueError('radius_tck must be a non-parametric spline')
    points = _get_points(tck, num_points)
    radii = _get_points(radius_tck, num_points=len(points))
    return points, radii



import cv2
import matplotlib.pyplot as plt
import numpy as np
from skimage.segmentation import active_contour
from skimage.draw import line, polygon, circle, ellipse


class ActiveContour(object):
    """docstring for ActiveContour

    Attributes:
        img (array): Input image
        mask (array): Input binary mask
        img_closing (): Image after Mrophology closing
        init_contour (): Initial contour
        mask_hole (): Mask of enclosed backgound area, which can be regard as
                      a hole in the mask. Active contour segmentation can't 
                      distinguish enclosed backgound area, but it can be obtained 
                      from input mask generated by GraphCut segmentation.
        snake (list): list of points of final contour
        seg_mask (): Segmentation mask
    """

    def __init__(self, img, mask, beta=10000):
        super(ActiveContour, self).__init__()
        self.img = img
        self.mask = mask

        # Pipeline:
        self.mroph_processing()
        self.find_init_contour()
        self.active_contour(beta)
        self.fill_contour()

    def mroph_processing(self, iters=1):
        kernel = np.ones((5, 5), np.uint8)
        img_closing = cv2.morphologyEx(self.mask+0, cv2.MORPH_CLOSE, kernel)
        self.img_closing = img_closing

    def find_init_contour(self):
        image, contours, hierarchy = cv2.findContours(
            self.img_closing, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        assert len(contours) == 1
        self.init_contour = contours[0]
        mask_cnt = np.zeros_like(self.mask)
        cv2.drawContours(mask_cnt, contours, 0, 255, -1)
        self.mask_hole = mask_cnt - self.img_closing

    def active_contour(self, beta=1000):
        init = [[cord[0][1], cord[0][0]] for cord in self.init_contour]
        init = np.array(init)
        self.snake = active_contour(self.img,
                                    init,
                                    beta=beta,
                                    w_edge=100,
                                    w_line=100,
                                    max_px_move=0.1,
                                    max_iterations=50,
                                    coordinates='rc')

    def fill_contour(self):
        self.seg_mask = np.zeros(self.mask.shape, np.uint8)
        rr, cc = polygon(self.snake[:, 0],
                         self.snake[:, 1], self.seg_mask.shape)
        self.seg_mask[rr, cc] = 255
        if cv2.countNonZero(self.mask_hole) > 100:
            self.seg_mask = self.seg_mask - self.mask_hole

    def plot(self):
        fig, ax = plt.subplots(figsize=(7, 7))
        ax.imshow(self.img, cmap=plt.cm.gray)
        # ax.plot(init[:, 1], init[:, 0], '--r', lw=3)
        ax.plot(self.snake[:, 1], self.snake[:, 0], '-b', lw=2)
        ax.plot(self.init_contour[:, :, 0],
                self.init_contour[:, :, 1], '--r', lw=2)
        plt.show()


def main():
    img = cv2.imread('1.jpg')
    img = cv2.resize(img, (400, 400))
    mask = cv2.imread('1_mask.png', 0)
    mask = cv2.resize(mask, (400, 400))
    active_contour = ActiveContour(img, mask)
    active_contour.plot()
    # cv2.imshow('img', active_contour.seg_mask)
    # cv2.waitKey(0)


if __name__ == '__main__':
    main()



import cv2
import numpy as np
import thinning
import networkx as nx
from itertools import combinations


class SkeletonExtractor(object):
    """Extract skeleton. First thinning, then make a graph to represent the branching
       structure. By searching longest path to find the path of skeleton

    Attributes:
        mask (array): Input segmentation mask
        thinned (array): Binary image. Skelton is 1-pixel width curve
        pixelpoints (list): List of points(y,x) on thinned image
        tip_pts (list): List of points that at end of curve
        fork_pts (list): List of points that connection multiple branches
        inter_pts (list): List of points that are in the middle of a branch
        graph (Graph): A graph representing the structure of branches on thinned image
        skeleton (list): Sequence of points on skeleton which do not have any branch
        mask_skeleton (array): Binary mask on which drawn the skeleton
    """

    def __init__(self, mask):
        super(SkeletonExtractor, self).__init__()
        self.mask = mask

        # Pipeline:
        self.thinning()
        self.classify_points()
        self.make_graph()
        self.find_skeleton()
        self.draw_skeleton()

    def thinning(self):
        self.thinned = thinning.guo_hall_thinning(self.mask+0)

    def classify_points(self):
        kernel = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]])
        dst = (cv2.filter2D(self.thinned/255, cv2.CV_64F, kernel))
        dst2 = np.uint8(dst)
        # only reserve pixels on thinned image
        dst2 = cv2.bitwise_and(dst2, dst2, mask=self.thinned)
        self.pixelpoints = np.transpose(np.nonzero(dst2))

        self.tip_pts = [[pt[0], pt[1]]
                        for pt in self.pixelpoints if dst2[pt[0]][pt[1]] == 1]
        self.fork_pts = [[pt[0], pt[1]]
                         for pt in self.pixelpoints if dst2[pt[0]][pt[1]] > 2]
        self.inter_pts = [[pt[0], pt[1]]
                          for pt in self.pixelpoints if dst2[pt[0]][pt[1]] == 2]

    def find_neighbors(self, x, y, mask):
        h, w = mask.shape
        neighbors = []
        for i in range(max(0, x-1), min(w, x+2)):
            for j in range(max(0, y-1), min(w, y+2)):
                if mask[j, i] > 0 and [i, j] != [x, y]:
                    neighbors.append([i, j])
        return neighbors

    def make_graph(self):
        self.graph = nx.Graph()
        self.graph.add_nodes_from(['%d,%d' % (pt[1], pt[0])
                                   for pt in self.pixelpoints])
        for y, x in self.pixelpoints:
            neighbors = self.find_neighbors(x, y, self.thinned)
            for i in neighbors:
                self.graph.add_edge('%d,%d' % (x, y), '%d,%d' % (i[0], i[1]))

    def find_skeleton(self):
        paths = []
        for i in combinations(self.tip_pts, 2):
            paths += nx.all_simple_paths(self.graph, source='%d,%d' %
                                         (i[0][1], i[0][0]), target='%d,%d' % (i[1][1], i[1][0]))
        self.skeleton = max(paths, key=len)
        self.skeleton = [[int(pt.split(',')[0]), int(
            pt.split(',')[1])] for pt in self.skeleton]

    def draw_skeleton(self):
        self.mask_skeleton = np.zeros(self.mask.shape)
        skeleton_expend_dim = np.expand_dims(self.skeleton, axis=1)
        cv2.drawContours(self.mask_skeleton, np.array(
            skeleton_expend_dim), -1, 255, 1)


if __name__ == '__main__':
    mask = cv2.imread('mask_0.png', 0)
    skeleton_extractor = SkeletonExtractor(mask)
    cv2.imshow('img', skeleton_extractor.mask_skeleton)
    cv2.waitKey(0)



import numpy
from scipy import ndimage
from . import spline_geometry
from . import interpolate


def sample_image_along_spline(image, tck, width, length=None, width_distance=None, order=3, **kwargs):
    """Return an image "swath" of a given length and width in pixels, centered
    along a parametric spline.

    Parameters:
    image: the image to sample from.
    tck: a parametric spline giving x, y coordinates to sample the image along.
    width: number of image samples to take perpendicular to the spline.
    length: number of image samples to take along the spline. For the resampling
        to neither magnify nor minify the image, this length ought to be the arc
        length of the spline. If more samples are taken than the arc length in
        pixels, then the image is magnified in the along-spline direction, and
        if fewer, then minification is obtained. If None, then the arc length
        is calculated and used.
    width_distance: distance (in image pixels) to sample perpendicular to the
        spline. If None, this defaults to width, resulting in no magnification
        or minification in that direction. If width_distance > width there will
        be minification, or if width_distance < width there will be
        magnification.
    order: image interpolation order for the resampling process.
        0 = nearest-neighbor interpolation
        1 = linear interpolation
        3 = cubic interpolation
    other keyword args are passed to ndimage.map_coordinates() to control the
        resampling. Useful arguments include 'mode' and 'cval' (see the
        documentation for ndimage.map_coordinates() for details.)

    Returns: array of shape (length, width) containing samples along the length
    of the spline, going perpendicular from the spline width_distance pixels
    on each side."""
    points, offset_directions = get_spline_sample_positions(tck, width, length, width_distance)
    sample_coordinates = points + offset_directions
    warped = ndimage.map_coordinates(image, sample_coordinates, order=order, **kwargs)
    return warped


def warp_image_to_standard_width(image, tck, src_radius_tck, dst_radius_tck, width, length=None, width_distance=None, order=3, **kwargs):
    """Return an image swath of a given length and width in pixels, centered
    along a parametric spline, and stretched non-uniformly perpendicular to
    the spline, such that an object in the image with an outline a determined
    by the spline and a radius-profile will be warped to a new radius-profile.

    This is most useful for stretching any given worm image to match some
    standardized "unit worm" in both length and radial profile.

    Parameters: image, tck, width, length, width_distance, order, kwargs are all
        as in sample_image_along_spline()
    src_radius_tck: nonparametric spline defining the raidus of the object along
        the tck centerline spline.
    dst_radius_tck: nonparametric spline defining the desired raidus profile.

    Returns: sampled array, as in sample_image_along_spline().
    """
    points, offset_directions = get_spline_sample_positions(tck, width, length, width_distance)
    length = points.shape[1]
    src_widths = interpolate.spline_interpolate(src_radius_tck, num_points=length)
    dst_widths = interpolate.spline_interpolate(dst_radius_tck, num_points=length)
    dst_widths[dst_widths == 0] = 1
    width_ratios = src_widths / dst_widths # shape = (length,)
    sample_coordinates = points + offset_directions * width_ratios[:, numpy.newaxis]
    # warped = ndimage.map_coordinates(image, sample_coordinates, order=order, **kwargs)
    warped = []
    for i in range(image.shape[2]): # To support 3 channel color image
        warped.append(
            ndimage.map_coordinates(
                image[:, :, i], sample_coordinates, order=order, **kwargs))
    warped = numpy.dstack(warped)
    return warped


def make_mask_for_sampled_spline(length, width, radius_tck, width_distance=None):
    """Given a the length and width of an image produced by sample_image_along_spline(),
    and a nonparametric spline giving the radius of an object (i.e. worm) along
    that spline, return a boolean mask that is True in the interior of that object.

    Parameters:
    length, width: the shape of the image produced by sample_image_along_spline()
        or similar.
    radius_tck: a nonparametric spline giving the radius (i.e. half-width) of
        the object at every point along the spline used for image sampling, in
        pixel units of the sampled image.
    width_distance: same as 'width_distance' parameter to sample_image_along_spline():
        this refers to the distance in source-image pixels sampled along the
        width of the output image. If None, assume this is the same as 'width'.

    Returns: boolean array of shape (length, width) that is True for every pixel
        that is within the radius of the object at each point along the image.
    """
    radius_at_each_pixel = interpolate.spline_interpolate(radius_tck, num_points=length)
    if width_distance is None:
        width_distance = width
    distance_from_image_centerline = numpy.abs(numpy.linspace(-width_distance/2, width_distance/2, width))
    # radius_at_each_pixel is 'length' pixels long,
    # distance_from_image_centerline is 'width' pixles wide
    # we want an image where element (i, j) is the result of
    # radius_at_each_pixel[i] > distance_from_image_centerline[j]
    # (i.e. 'True' where for all pixels that are closer to the centerline than
    # the edge of the object). Some thinking should convince you that this is
    # the "outer product" of the greater-than operator...
    return numpy.greater.outer(radius_at_each_pixel, distance_from_image_centerline)


def sample_image_rect(image, size, center, rotation=0, order=3, **kwargs):
    """Return an image "swath" of a given (w, h) size, centered at a given (x, y)
    position and rotated by the specified degrees.

    Parameters:
    image: the image to sample from.
    size: (w, h) in pixels of the output image
    center: (x, y) position in the input image that the sampling region is
        centered on.
    rotation: rotation in degrees to apply to the sampling region.
        NB: for images, the positive x axis goes downward on the screen. So
        the rotation direction seems negated compared to conventional
        coordinates. (That is, a slight negative rotation will cause the
        sampling rect to be rotated a little counterclockwise.)
    order: image interpolation order for the resampling process.
        0 = nearest-neighbor interpolation
        1 = linear interpolation
        3 = cubic interpolation
    other keyword args are passed to ndimage.map_coordinates() to control the
        resampling. Useful arguments include 'mode' and 'cval' (see the
        documentaiton for ndimage.map_coordinates() for details.)

    Returns: array of the specified size containing the image pixels sampled
    from the region as defined by the parameters above."""

    size = numpy.asarray(size)
    x, y = numpy.indices(size) - (size / 2).reshape((2,1,1))
    theta = numpy.pi * rotation / 180
    rx = x*numpy.cos(theta)-y*numpy.sin(theta)
    ry = x*numpy.sin(theta)+y*numpy.cos(theta)
    coords = numpy.array([rx, ry]) + numpy.reshape(center, (2,1,1))
    sampled = ndimage.map_coordinates(image, coords, order=order, **kwargs)
    return sampled


def get_spline_sample_positions(tck, width, length=None, width_distance=None):
    """Calculate positions along the spline and vector offsets from those points
    for each sample position along the perpendiculars to that spline.

    See the documentation for sample_image_along_spline() for explanation of the
    parameters.

    Returns: points, offset_directions
        points.shape = (2, length, 1)
        offset_directions.shape = (2, length, width)
        (Note: points's shape is such that it can simply be added to the
        offset_directions array, via numpy broadcasting.)
    """
    if length is None:
        length = int(round(spline_geometry.arc_length(tck)))
    if width_distance is None:
        width_distance = width

    perpendiculars = spline_geometry.perpendiculars(tck, length).T # shape = (2, length)
    offsets = numpy.linspace(-width_distance/2, width_distance/2, width) # distances along each perpendicular across the width of the sample swath
    offset_directions = numpy.multiply.outer(perpendiculars, offsets) # shape = (2, length, width)
    points = interpolate.spline_interpolate(tck, length) # shape = (length, 2)
    points = points.T[..., numpy.newaxis] # now points.shape = (2, length, 1)
    return points, offset_directions



import cv2
import matplotlib.pyplot as plt
import numpy as np
import rel_imp
if not __name__ == '__main__':
    from .utils import geometry, interpolate, spline_geometry, resample


class Straightener(object):
    """Straighten a warped object by resampling along skeleton

    Attributes:
        img (array): Input image
        skeleton (list): Point sequence of skeleton which have no branch
        widths (list): Local width at each position along skeleton
        img_straightened: Result image after straightening
        mask_straightened: Mask of valid area in straightened image
    """

    def __init__(self, img, skeleton, widths):
        super(Straightener, self).__init__()
        self.img = img
        self.skeleton = skeleton  # skeleton is a point sequence(list)
        self.widths = widths

        # Pipeline:
        self.fit_spline()
        self.straighten()

    def fit_spline(self):
        # sample_widths are the widths between skeleton and contour
        sample_widths = [width/1.8+5 for width in self.widths]
        self.spine_tck = interpolate.fit_spline(
            [[y, x] for x, y in self.skeleton], smoothing=None)
        x_vals = np.linspace(0, 1, len(self.skeleton))
        self.width_tck = interpolate.fit_nonparametric_spline(
            x_vals, sample_widths, smoothing=None)

    def straighten(self):
        overlay_width = int(max(self.widths) * 1.2)
        self.img_straightened = resample.warp_image_to_standard_width(
            self.img,
            self.spine_tck,
            self.width_tck,
            self.width_tck,
            overlay_width)
        self.mask_straightened = resample.make_mask_for_sampled_spline(
            self.img_straightened.shape[0],
            self.img_straightened.shape[1],
            self.width_tck)
        self.img_straightened[~self.mask_straightened] = 0
        self.img_straightened = cv2.rotate(
            self.img_straightened, cv2.ROTATE_90_COUNTERCLOCKWISE)
        self.mask_straightened = self.mask_straightened.transpose()[::-1]
        return self.img_straightened


if __name__ == '__main__':
    from skeleton_extractor import SkeletonExtractor
    from pixel_annotator import PixelAnnotator
    from utils import geometry, interpolate, spline_geometry, resample
    path = '1.jpg'
    img = cv2.imread(path)
    mask = cv2.imread('mask_0.png', 0)
    _, mask = cv2.threshold(mask, 127, 255, cv2.THRESH_BINARY)
    skeleton_extractor = SkeletonExtractor(mask)
    skeleton = skeleton_extractor.skeleton
    pixel_annotator = PixelAnnotator(mask, skeleton)
    staightener = Straightener(img, skeleton, pixel_annotator.widths)
    cv2.imshow('img', staightener.img_straightened)
    cv2.waitKey(0)



import cv2
import matplotlib.pyplot as plt
import numpy as np


class EdgeEnhancing(object):
    """Enhancing edges by adding shadow
    
    Attributes:
        img (array): Input image
        img_gray (array): Gray scale image of input image
        img_denoise (array): Denoised image of img_gray
        img_edge_enhanced (array): Image with shadow on img_denoise
    """

    def __init__(self, img=None):
        super(EdgeEnhancing, self).__init__()
        self.img = img
        self.img_gray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)

        # Pipeline:
        self.remove_overexposure()
        self.denoise()
        self.add_shadow()

    def remove_overexposure(self):
        self.img_gray[self.img_gray == 255] = 127

    def denoise(self):
        self.img_denoise = cv2.fastNlMeansDenoising(
            self.img_gray, None, 10, 7, 21)

    def add_shadow(self):
        emboss_kernels = {
            'north_east':  [[0, 0, 1],
                            [0, 0, 0],
                            [-1, 0, 0]],
            'south_west': [[0, -1, -1],
                           [1, 0, -1],
                           [1, 1, 0]],
            'south_east': [[-1, -1, 0],
                           [-1, 0, 1],
                           [0, 1, 1]],
            'north_west': [[1, 0, 0],
                           [0, 0, 0],
                           [0, 0, -1]],
            'north':      [[1, 2, 1],
                           [0, 0, 0],
                           [-1, -2, -1]],
            'south':      [[-1, -2, -1],
                           [0, 0, 0],
                           [1, 2, 1]],
            'east':       [[-1, 0, 1],
                           [-2, 0, 2],
                           [-1, 0, 1]],
            'west':       [[1, 0, -1],
                           [2, 0, -2],
                           [1, 0, -1]],
        }

        self.img_edge_enhanced = self.img_denoise + 0
        for direction in emboss_kernels:
            kernel = np.array(emboss_kernels[direction])
            shadow = cv2.filter2D(self.img_denoise, -1, kernel)
            self.img_edge_enhanced = cv2.subtract(self.img_edge_enhanced, shadow)


if __name__ == '__main__':
    path = '1.jpg'
    img = cv2.imread(path)
    edge_enhancing = EdgeEnhancing(img)

    cv2.imshow('img', edge_enhancing.img_edge_enhanced)
    cv2.waitKey(0)
    cv2.imwrite('1_edge_enhanced.png',edge_enhancing.img_edge_enhanced)



import numpy
from scipy.interpolate import _fitpack_impl as fitpack
from . import geometry

def linear_resample_polyline(points, num_points):
    """Resample a piecewise linear curve to contain a given number of
    equally-spaced points, using linear interpolation.

    Parameters:
    points: array of n points x,y; shape=(n,2)
    num_points: number of output points in array.

    Returns a resampled array, of shape (num_points,2)"""
    points = numpy.asarray(points)
    distances = geometry.cumulative_distances(points, unit=True)
    sample_positions = numpy.linspace(0, 1, num_points)
    x = numpy.interp(sample_positions, distances, points[:,0])
    y = numpy.interp(sample_positions, distances, points[:,1])
    return numpy.transpose([x,y])


def spline_resample_polyline(points, num_points):
    """Resample a piecewise linear curve to contain a given number of
    equally-spaced points, using spline interpolation with automatically calculated
    smoothing.

    Parameters:
    points: array of n points x,y; shape=(n,2)
    num_points: number of output points in array.

    Returns a resampled array, of shape (num_points,2), and the spline parameters
    used for the resampling"""
    tck = fit_splines(points)
    points_out = spline_interpolate(tck, num_points)
    return points_out, tck


def fit_spline(points, smoothing=None, order=None, force_endpoints=True):
    """Fit a parametric smoothing spline to a given set of x,y points. (Fits
    x(p) and y(p) as functions for some parameter p.)

    Parameters:
    points: array of n points x,y; shape=(n,2)
    smoothing: smoothing factor: 0 requires perfect interpolation of the
        input points, at the cost of potentially high noise. Very large values
        will result in a low-order polynomial fit to the points. If None, an
        appropriate value based on the scale of the points will be selected.
        Specifically, the fitting function guarantees that sum of the distances
        between the input points and the spline will be less than or equal to
        this smoothing parameter.
    order: The desired order of the spline. If None, will be 1 if there are
        three or fewer input points, and otherwise 3.
    force_endpoints: if True (default), the endpoints of the spline will be set
        to match the positions of the input data, regardless of smoothing.
        NB: With large smoothing, this can dramatically influence the position
        of the entire curve. With smaller smoothing, changing the endpoints
        will have a more local effect.

    Returns a spline tuple (t,c,k) consisting of:
        t: the knots of the spline curve
        c: the x and y b-spline coefficients for each knot (shape (m, 2))
        k: the order of the spline.

    Note: the smoothing factor is an upper bound on the sum of all the squared
    distances between the original x,y points and the matching points on the
    smoothed spline representation."""
    points = numpy.asarray(points)
    l = len(points)
    w = numpy.ones(l, float)
    # choose input parameter values for the curve as the distances along the polyline:
    # this gives something close to the "natural parameterization" of the curve.
    # (i.e. a parametric curve with first-derivative close to unit magnitude: the curve
    # doesn't accelerate/decelerate, so points along the curve in the x,y plane
    # don't "bunch up" with evenly-spaced parameter values.)
    distances = geometry.cumulative_distances(points, unit=False)
    if numpy.any(numpy.isclose(distances[1:] - distances[:-1], 0)):
        raise ValueError("Repeated input points are not allowed.")
    if order is None:
        if l < 4:
            k = 1
        else:
            k = 3
    else:
        k = order
    if smoothing is None:
        smoothing = l * distances[-1] / 600.
    if force_endpoints:
        w[[0,-1]] = l

    t, c, ier, msg = splprep(distances, points, s=smoothing, k=k, w=w)

    if ier > 3:
        raise RuntimeError(msg)
    if force_endpoints:
        # endpoints should already be close, so this shouldn't distort the spline too much...
        c[[0,-1]] = points[[0,-1]]

    return t, c, k

def fit_nonparametric_spline(x, y, smoothing=None, order=None, force_endpoints=True):
    """Fit a non-parametric smoothing spline to x,y points. (Fits a function
    f(x) = y, or approximately so.)

    Parameters:
    x: array of shape=(n,). Must be monotonic
    y: array of shape=(n,)
    smoothing: smoothing factor: 0 requires perfect interpolation of the
        input points, at the cost of potentially high noise. Very large values
        will result in a low-order polynomial fit to the points. If None, an
        appropriate value based on the scale of the points will be selected.
        Specifically, the fitting function guarantees that sum of the distances
        between the input y values and the spline at the same x values will be
        less than or equal to this smoothing parameter.
    order: The desired order of the spline. If None, will be 1 if there are
        three or fewer input points, and otherwise 3.
    force_endpoints: if True (default), the endpoints of the spline will be set
        to match the positions of the input data, regardless of smoothing.
        NB: With large smoothing, this can dramatically influence the position
        of the entire curve. With smaller smoothing, changing the endpoints
        will have a more local effect.

    Returns a spline tuple (t,c,k) consisting of:
        t: the knots of the spline curve
        c: the b-spline coefficient for each knot
        k: the order of the spline.

    Note: the smoothing factor is an upper bound on the sum of all the squared
    distances between the original y values and the matching points on the
    smoothed spline representation."""

    x = numpy.asarray(x)
    y = numpy.asarray(y)
    l = len(x)
    w = numpy.ones(l, float)

    if order is None:
        if l < 4:
            k = 1
        else:
            k = 3
    else:
        k = order
    if smoothing is None:
        smoothing = l * abs(x[0] - x[-1]) / 600.
    if force_endpoints:
        w[[0,-1]] = l

    t, c, ier, msg = splrep(x, y, s=smoothing, k=k, w=w)

    if ier > 3:
        raise RuntimeError(msg)
    if force_endpoints:
        # endpoints should already be close, so this shouldn't distort the spline too much...
        c[[0,-1]] = y[[0,-1]]

    return t, c, k


def spline_interpolate(tck, num_points, derivative=0):
    """Return num_points equally spaced along the given spline.

    If derivative=0, then the points themselves will be given; if derivative>0
    then the derivatives at those points will be returned."""
    # t[-1] gives the maximum value for the input position
    output_positions = numpy.linspace(0, tck[0][-1], num_points)
    return spline_evaluate(tck, output_positions, derivative)


def spline_evaluate(tck, positions, derivative=0):
    """Evaluate a spline at the given positions.

    If derivative=0, then the points themselves will be given; if derivative>0
    then the derivatives at those points will be returned."""
    t, c, k = tck
    if c.ndim == 1:
        evaluate = splev # non-parametric
    else:
        evaluate = splpev # parametric
    return evaluate(positions, t, c, k, der=derivative)


def reverse_spline(tck):
    """Reverse the direction of a spline (parametric or nonparametric),
    without changing the range of the t (parametric) or x (nonparametric) values."""
    t, c, k = tck
    rt = t[-1] - t[::-1]
    rc = c[::-1]
    return rt, rc, k


def insert_control_points(tck, num_points):
    """Return an equivalent spline with additional control points added for
    improved editability.

    Parameters:
    tck: spline tuple
    num_points: total number of spline knots in returned spline tuple

    Returns: a new tck tuple"""
    t, c, k = tck
    if len(c.shape) > 1:
        # parametric spline: use pinsert
        insert_func = pinsert
    else:
        # non-parametric: use normal insert
        insert_func = insert
    while len(t) < num_points:
        spans = t[1:] - t[:-1]
        p = spans.argmax()
        new_t = t[p:p+2].mean()
        t, c, k = insert_func(t, c, k, new_t)
    return t, c, k


def spline_to_bezier(tck):
    """Convert a parametric spline into a sequence of Bezier curves of the same degree.

    Returns a list of Bezier curves of degree k that is equivalent to the input spline.

    Each Bezier curve is an array of shape (k+1,d) where d is the dimension of the
    space; thus the curve includes the starting point, the k-1 internal control
    points, and the endpoint, where each point is of d dimensions."""
    t, c, k = tck
    t = numpy.asarray(t)
    old_t = t
    try:
        c[0][0]
    except:
        # I can't figure out a simple way to convert nonparametric splines to
        # parametric splines. Oh well.
        raise TypeError("Only parametric splines are supported.")
    # the first and last k+1 knots are identical in the non-periodic case, so
    # no need to consider them when increasing the knot multiplicities below
    knots_to_consider = numpy.unique(t[k+1:-k-1])
    # For each unique knot, bring its multiplicity up to the next multiple of k+1
    # This removes all continuity constraints between each of the original knots,
    # creating a set of independent Bezier curves.
    desired_multiplicity = k+1
    for x in knots_to_consider:
        current_multiplicity = numpy.sum(old_t == x)
        remainder = current_multiplicity%desired_multiplicity
        if remainder != 0:
            # add enough knots to bring the current multiplicity up to the desired multiplicity
            number_to_insert = desired_multiplicity - remainder
            t, c, k = pinsert(t, c, k, x, number_to_insert)
    # group the points into the desired bezier curves
    return numpy.split(c, len(c) // desired_multiplicity)


def pinsert(t, c, k, u, m=1):
    """Insert m control points in spline (t,c,k) at parametric position u."""
    c = c.T
    out = numpy.empty((c.shape[1]+m, c.shape[0]))
    per=False
    for i, cc in enumerate(c):
        tt, ccc, ier = fitpack._fitpack._insert(per, t, cc, k, u, m)
        out[:,i] = ccc[:-k-1]
        if ier==10: raise ValueError("Invalid input data")
        if ier: raise TypeError("An error occurred")
    return tt, out, k


def insert(t, c, k, x, m=1):
    """Insert m control points in spline (t,c,k) at parametric position x."""
    per=False
    t, c, ier = fitpack._fitpack._insert(per, t, c, k, x, m)
    if ier==10: raise ValueError("Invalid input data")
    if ier: raise TypeError("An error occurred")
    return t, c, k


def splrep(x, y, s, k, w=None):
    """Return degree-k spline representation (t,c,k) of x, y points, with smoothing parameter s and weights w.

    Smoothing parameter guarantee:
    ((w * (splev(x, *tck) - y))**2).sum() <= s

    """
    m = x.shape[0]
    if w is None:
        w = numpy.ones(m, float)
    xb, xe = x[[0, -1]]
    if not (1 <= k <= 5): raise TypeError('1<=k=%d<=5 must hold'%(k))
    if (m != len(y)):
            raise TypeError('Lengths of the first two must be equal')
    if m <= k: raise TypeError('m>k must hold')
    nest = m + k + 1
    t = numpy.empty((nest,), float)
    wrk = numpy.empty((m*(k+1)+nest*(7+3*k),), float)
    iwrk = numpy.empty((nest,), numpy.int32)
    task = 0
    n, c, fp, ier = fitpack.dfitpack.curfit(task, x, y, w, t, wrk, iwrk, xb, xe, k, s)
    return t[:n], c[:n-k-1], ier, fitpack._iermess[ier][0]

def splprep(u, x, s, k, w=None):
    """Return spline representation (t,c,k) of parametric curve x(u), with smoothing parameter s and weights w.

    Parameters:
    u: array of shape (n) containing parametric positions
    x: array of shape (n, m) containing n points in m dimensions, which are the positions of the
       curve x(u) at each parametric value in the array u.
    s: smoothing parameter (see below)
    k: degree of output spline
    w: weights for each point.

    Returns spline tuple (t,c,k)

    Smoothing parameter guarantee:
    ((w * (splpev(u, *tck) - x))**2).sum() <= s"""

    m, idim = x.shape
    if w is None:
        w = numpy.ones(m, float)
    ub, ue = u[[0, -1]]
    if not (1 <= k <= 5): raise TypeError('1<=k=%d<=5 must hold'%(k))
    if not len(u) == m:
            raise TypeError('Mismatch of input dimensions')
    if m <= k: raise TypeError('m>k must hold')
    nest = m + k + 1
    t = numpy.array([], float)
    wrk = numpy.array([], float)
    iwrk = numpy.array([], numpy.int32)
    task = per = 0
    ipar = True
    t,c,o = fitpack._fitpack._parcur(x.ravel(), w, u, ub, ue, k, task, ipar, s, t, nest, wrk, iwrk, per)
    ier, fp, n = o['ier'], o['fp'], len(t)
    c.shape = (idim, n-k-1)
    return t, c.T, ier, fitpack._iermess[ier][0]


def splev(x, t, c, k, der=0):
    "Evaluate spline (t,c,k) or nth-order derivative thereof at position x."
    c = c.T
    out, ier = fitpack._fitpack._spl_(x, der, t, c, k, 0)
    if ier==10: raise ValueError("Invalid input data")
    if ier: raise TypeError("An error occurred")
    return out


def splpev(u, t, c, k, der=0):
    "Evaluate parametric spline (t,c,k) or nth-order derivative thereof at position u."

    c = c.T
    out = numpy.empty((len(u), c.shape[0]))
    for i, cc in enumerate(c):
        out[:,i], ier = fitpack._fitpack._spl_(u, der, t, cc, k, 0)
        if ier==10: raise ValueError("Invalid input data")
        if ier: raise TypeError("An error occurred")
    return out




import cv2
import matplotlib.pyplot as plt
import numpy as np
from skimage import data, filters, segmentation, measure, morphology, color, io
import thinning


class ROIFinder(object):
    """Find Eligible ROIs in a segmentation mask

    Attributes:
        mask (array): Input binary mask
        area_threshold (): Threshold of min area of an ROI
        markers (array): Label each pixel by the connected component it belongs
        valid_labels (list): Labels of Eligible ROIs
        valid_regions (list): Eligible ROIs

    Usage:
        yield_roi(): Return mask of a eligible ROI
    """

    def __init__(self, mask, area_threshold=200):
        super(ROIFinder, self).__init__()
        self.mask = mask
        self.area_threshold = area_threshold

        self.checklist = [
            self.check_area,
            self.check_extent,
            self.check_not_blob,
            self.check_shape,
        ]

        # Pipeline:
        self.mark()
        self.filter()

    def mark(self):
        self.markers = measure.label(self.mask)
        # self.image_label_overlay =color.label2rgb(self.markers, image=self.mask,bg_label=0)

    def check_area(self, region):  # prevent region is too small
        return region.area > self.area_threshold

    def check_extent(self, region):  # prevent region is a rectangle
        return region.extent < 0.8

    def check_not_blob(self, region):  # prevent region is a circular blob
        mask_region = np.zeros_like(self.mask)
        mask_region[self.markers == region.label] = 255
        image, contours, hierarchy = cv2.findContours(
            mask_region, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        (x, y), radius = cv2.minEnclosingCircle(contours[0])
        circle_area = np.pi*radius**2
        return region.area/circle_area < 0.5

    def check_shape(self, region):
        mask_region = np.zeros_like(self.mask)
        mask_region[self.markers == region.label] = 255
        kernel = np.ones((5, 5), np.uint8)
        mask_region = cv2.dilate(mask_region, kernel, iterations=1)
        thinned = thinning.guo_hall_thinning(mask_region+0)
        length = cv2.countNonZero(thinned)
        ratio = region.area/length
        return ratio > 10 and length > 50

    def is_valid_region(self, region):
        for check in self.checklist:  # Return False if any condition not satisfied
            if not check(region):
                return False
        return True

    def filter(self):
        self.valid_labels = []
        self.valid_regions = []
        for i, region in enumerate(measure.regionprops(self.markers)):
            if self.is_valid_region(region):
                self.valid_labels.append(region.label)
                self.valid_regions.append(region)

    def yield_roi(self):  # Return a generator
        i = len(self.valid_regions)
        while i > 0:
            roi_mask = np.zeros(self.mask.shape, np.uint8)
            roi_mask[self.markers == self.valid_labels[i-1]] = 255
            i -= 1
            yield roi_mask


if __name__ == '__main__':
    mask = cv2.imread('1_mask.png', 0)
    roi_finder = ROIFinder(mask)
    for roi in roi_finder.yield_roi():
        cv2.imshow('img', roi)
        cv2.waitKey(0)



from .active_contour import ActiveContour
from .edge_enhancing import EdgeEnhancing
from .graph_cut import GraphCut
from .pixel_annotator import PixelAnnotator
from .skeleton_extractor import SkeletonExtractor
from .straightener import Straightener
from .roi_finder import ROIFinder




import numpy

def cumulative_distances(points, unit=True):
    """Return cumulative distances along a polyline.

    Parameters:
    points: array of shape (n,m) consisting of n points in m dimensions
    unit: if True, return distances divided by total length of the curve,
          if False, return actual arc lengths."""
    points = numpy.asarray(points)
    distances = numpy.concatenate([[0], numpy.add.accumulate(numpy.sqrt(((points[:-1] - points[1:])**2).sum(axis=1)))])
    if unit:
        distances /= distances[-1]
    return distances

def filter_dup_points(points):
    """Return a polyline with no duplicate or near-duplicate points."""
    points_out = [points[0]]
    for point in points[1:]:
        if not numpy.allclose(point, points_out[-1]):
            points_out.append(point)
    return numpy.array(points_out)

def closest_point(point, points):
    """Find the closest position in in array 'points' to the provided 'point'."""
    points_arr = numpy.asarray(points)
    distances_squared = ((points_arr - point)**2).sum(axis=1)
    i = numpy.argmin(distances_squared)
    return i, points[i]

def closest_point_to_line_segments(point, lines_start, lines_end):
    """Given a point and a set of line segments (specified by starting
    and ending points), return the point on each line segment that is closest to the
    given point, and the parametric position along each line of that point."""
    v = lines_end - lines_start
    w = point - lines_start
    c1 = (v*w).sum(axis=1)
    c2 = (v*v).sum(axis=1)
    fractional_positions = c1 / c2
    fractional_positions = fractional_positions.clip(0, 1)
    closest_points = lines_start + fractional_positions[:,numpy.newaxis]*v
    return closest_points, fractional_positions

def closest_point_on_polyline(point, points, parameters=None):
    """Return the point along a polyline nearest the given point and the parametric
    position of that point along the polyline. If no input parameter values are
    given, then the cumulative distance along the polyline will be taken as the parameter."""
    points = numpy.asarray(points)
    closest_points, fractions = closest_point_to_line_segments(point, points[:-1], points[1:])
    distances = numpy.sqrt(((point - closest_points)**2).sum(axis=1))
    point_idx = distances.argmin()
    closest_point = closest_points[point_idx]
    if parameters is None:
        parameters = cumulative_distances(spine)
    start_u, stop_u = parameters[point_idx:point_idx+2]
    u_val = start_u + fractions[point_idx]*(stop_u - start_u)
    return closest_point, u_val

def angle_between_vectors(v_from, v_to):
    """Calculate the angle in radians between two 2d vectors."""
    return numpy.arctan2(v_from[:,0]*v_to[:,1]-v_from[:,1]*v_to[:,0], (v_from * v_to).sum(axis=1))

def find_perp(p0, p1, unit=True):
    """Return a perpendicular to line p0-p1, optionally of unit-length.

    Parameters:
    p0 and p1 can be arrays of shape (m) each representing a single point in m dimensions,
    or of (n,m) containing n points in m dimensions. The returned array is either a
    array of shape (m) containing one or of shape (n,m) contaning n perpendiculars."""
    p1 = p1.astype(float)
    diff = p1 - p0
    diff = numpy.roll(diff, 1, axis=-1)
    diff[...,0] *= -1
    if unit:
        diff /= numpy.sqrt(numpy.sum(diff**2, axis=-1))[...,numpy.newaxis]
    return diff

def find_bisector(p0, p1, p2):
    """Find offset vector from p1 that bisects the angle created by three
    2d points, p0, p1, and p2."""
    d1 = p1 - p0
    d2 = p2 - p1
    a1 = numpy.arctan2(d1[...,1], d1[...,0])
    a2 = numpy.arctan2(d2[...,1], d2[...,0])
    ad = a2 - a1
    af = a1 + (numpy.pi + ad) / 2
    return numpy.transpose([numpy.cos(af), numpy.sin(af)])

def find_polyline_perpendiculars(points):
    """Find perpendiculars to an input polyline of shape (n, 2).
    At the endpoints, the perpendiculars are at right angles to the starting
    and ending line segments. At internal points, the perpendiculars bisect
    the angle of the internal positions.
    Note: special care is taken to make sure the perpindiculars point in a
    consistent direction at the internal positions, based on the overall direction
    of the polyline."""
    perpendiculars = numpy.empty(points.shape, dtype=float)
    perpendiculars[[0,-1]] = find_perp(points[[0, -2]], points[[1, -1]])
    bisectors = find_bisector(points[:-2], points[1:-1], points[2:])
    perps = find_perp(points[:-2], points[2:])
    dots = (bisectors * perps).sum(axis=1)
    perpendiculars[1:-1] = bisectors * numpy.sign(dots)[..., numpy.newaxis]
    return perpendiculars



import cv2
import matplotlib.pyplot as plt
import numpy as np


class GraphCut(object):
    """Segmentation using GraphCut algorithm

    Attributes:
        img (array): Input image
        img_gray: Gray scale image of input image
        img_enhanced: Image of which edges is enhanced by adding shadow
        img_denoise: Image after denoise
        img_corrected (array): Image after illumination correction
        rough_seg_mask: Segmentation mask by Graphcut, edges as initial label
        mask_fuzzy_zone: Mask of fuzzy zone in image
        delicate_seg_mask: Description
        bgdModel: Backgroud model of GraphCut
        fgdModel: Foreground model of GraphCut
        GRAPHCUT_ITER_TIMES (int): Iteration times when applying GraphCUt
        INPAINT_RADIUS (int): Inpaint color is decided by the color of ajacent area
                              within a circle of which radius is INPAINT_RADIUS
        edges: Edges extracted by Canny operator, used as initial label for GraphCut
        init_label: Initial label for GraphCut
        img_enhance_fuzzy_zone: Image of which fuzzy zone is enhanced

    """

    def __init__(self, img=None, img_enhanced=None,
                 illumination_correction=False, delicate=True):
        super(GraphCut, self).__init__()
        self.img = img
        self.img_gray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
        self.img_enhanced = self.img_gray if img_enhanced is None else img_enhanced
        self.img_denoise = cv2.fastNlMeansDenoising(
            self.img_gray, None, 7, 21, 21)
        self.img_corrected = self.img_denoise
        self.rough_seg_mask = np.zeros_like(self.img_gray)
        self.mask_fuzzy_zone = np.zeros_like(self.img_gray)
        self.delicate_seg_mask = np.zeros_like(self.img_gray)
        self.bgdModel = np.zeros((1, 65), np.float64)
        self.fgdModel = np.zeros((1, 65), np.float64)

        self.GRAPHCUT_ITER_TIMES = 1
        self.INPAINT_RADIUS = 30

        # Pipeline：
        self.init_labelling()
        self.rough_segment()
        if illumination_correction:
            self.illumination_correction()
        if delicate:
            self.pick_up_fuzzy_zone()
            self.delicate_segment()
            self.seg_mask = self.delicate_seg_mask
        else:
            self.seg_mask = self.rough_seg_mask

    def init_labelling(self):
        assert len(self.img_gray.shape) == 2
        self.edges = cv2.Canny(self.img_enhanced, 30, 150)

        self.init_label = np.zeros_like(self.img_gray)
        self.init_label[self.edges == 0] = 2  # 2 is PR_BGD
        self.init_label[self.edges == 255] = 3  # 3 is PR_FGD

    def rough_segment(self):
        label = self.init_label + 0
        cv2.grabCut(self.img, label, None, self.bgdModel, self.fgdModel,
                    self.GRAPHCUT_ITER_TIMES, cv2.GC_INIT_WITH_MASK)
        self.rough_seg_mask[label == 1] = 255
        self.rough_seg_mask[label == 3] = 255

    def illumination_correction(self):
        # Step 1: mask out objects
        kernel = np.ones((5, 5), np.uint8)
        mask_dilated = cv2.dilate(self.rough_seg_mask, kernel, iterations=2)
        img_inpaint = cv2.inpaint(self.img_gray, mask_dilated,
                                  self.INPAINT_RADIUS, cv2.INPAINT_NS)

        # Step 2: calculate light field and correct illumination
        blur_radius = max(self.img_gray.shape)//10
        blur = cv2.blur(img_inpaint, (blur_radius, blur_radius))
        mean = cv2.mean(blur)
        light_field = blur/mean[0]
        self.img_corrected = (self.img_denoise/light_field).astype(np.uint8)

    def pick_up_fuzzy_zone(self):
        backgournd_intensity = cv2.mean(self.img_gray,
                                        mask=255-self.rough_seg_mask)[0]
        img_backgroud = self.img_corrected+0
        img_backgroud[self.rough_seg_mask == 255] = int(backgournd_intensity)

        equ = cv2.equalizeHist(img_backgroud+0)
        ret2, th2 = cv2.threshold(equ, 30, 255, cv2.THRESH_BINARY)
        self.mask_fuzzy_zone[th2 == 0] = 255
        self.img_enhance_fuzzy_zone = cv2.addWeighted(
            equ, 0.5, self.img_denoise, 0.5, 0)

    def delicate_segment(self):
        label = np.zeros_like(self.img_gray)
        label[label == 0] = 2  # 2 is PR_BGD
        label[self.edges == 255] = 3  # 3 is PR_FGD
        label[self.mask_fuzzy_zone == 255] = 3  # 3 is PR_FGD

        img = cv2.cvtColor(self.img_enhance_fuzzy_zone, cv2.COLOR_GRAY2BGR)
        cv2.grabCut(img, label, None, self.bgdModel, self.fgdModel,
                    self.GRAPHCUT_ITER_TIMES, cv2.GC_INIT_WITH_MASK)
        self.delicate_seg_mask[label == 1] = 255
        self.delicate_seg_mask[label == 3] = 255


if __name__ == '__main__':
    path1 = '1.jpg'
    path2 = '1_edge_enhanced.png'
    img = cv2.imread(path1)
    img_enhanced = cv2.imread(path2, 0)
    graph_cut = GraphCut(img=img, img_enhanced=img_enhanced,
                         illumination_correction=False)
    cv2.imshow('img', graph_cut.delicate_seg_mask)
    cv2.waitKey(0)
    cv2.imwrite('1_mask.png', graph_cut.delicate_seg_mask)



#!/usr/bin/env python
# coding: utf-8

import setuptools

setuptools.setup(
    name='deepworm_algorithm',
    version='0.1.8',
    author='JohnsonNie',
    author_email='pubrcv@163.com',
    url='http://www.deepdiy.net/',
    description='A set of algorithms for analysing worm image',
    packages=setuptools.find_packages(),
    install_requires=[
        'thinning-py3',
        'opencv-python',
        'matplotlib',
        'scikit-image',
        'networkx',
        'scipy',
        'numpy', 
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import cv2
import numpy as np
import networkx as nx
from itertools import combinations
from scipy import spatial
import networkx as nx
import matplotlib.pyplot as plt


class PixelAnnotator(object):
    """Annotating a segmentation mask by labelling the mapping relationship 
       between points

    Attributes:
        mask (array): Input segmentation mask
        skeleton (list): Sequence of points of skeleton which have no branches
        mask_labeled_skeleton (array): A mask on which points are labele with its
            index on skeleton sequence
        cnt (list): List of points of contour. Has a redundent dimention[0]
        gradient_x (array): Gradient derived from Scharr operator
        gradient_y (array): Gradient derived from Scharr operator
        curvature_heatmap (array): curverate on each pixel
        end_pt_1 (list): Coordinates(x,y) of one end of skeleton
        end_pt_2 (list): Coordinates(x,y) of another end of skeleton
        contour_graph (Graph): A graph representing the ring structure of contour
        side1 (list): Sequence of points(x,y) of one side on the contour
        side2 (list): Sequence of points(x,y) of another side on the contour
        labels_side1 (list): Labels of corresponding skeleton points of points on side1
        labels_side2 (list): Labels of corresponding skeleton points of points on side2
        widths (list): Local width at each position along skeleton
        side1_widths (list): Local width between skeleton and side1 at each position 
                             along skeleton
        side2_widths (list): Local width between skeleton and side2 at each position 
                             along skeleton
        is_head_ahead (bool): Is the skeleton in the order of head->tail 
        keypoint_dict (dict): Dict of skeleton,two sides, labels of two sides
    """

    def __init__(self, mask, skeleton):
        super(PixelAnnotator, self).__init__()
        self.mask = mask
        self.skeleton = skeleton  # skeleton is a point sequence(list)
        self.mask_labeled_skeleton = np.zeros(mask.shape)

        # Fraction of head/tail length,for estimating body orientation
        self.integral_span = 0.15  # Head(tail) length over body length

        # Pipeline:
        self.find_contour()
        self.label_skeleton()
        self.map_contour_to_skeleton()  # map outline to skeleton
        self.compute_gradients()  # Gradient derived from Scharr operator
        self.draw_curvature_heatmap()
        self.find_end_points()
        self.make_contour_graph()  # graph of the ring structure of contour
        self.cut_graph_to_2_sides()
        self.map_skeleton_to_2_sides()  # map each side of outline to skeleton
        self.compute_widths()
        self.estimate_body_orientation()

    def find_contour(self):
        image, contours, hierarchy = cv2.findContours(
            self.mask+0, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
        assert len(contours) == 1
        self.cnt = contours[0].squeeze()

    def label_skeleton(self):
        for i, pt in enumerate(self.skeleton):
            self.mask_labeled_skeleton[pt[1], pt[0]] = i

    def map_contour_to_skeleton(self):
        tree = spatial.KDTree(np.array(self.skeleton))
        labels = tree.query(self.cnt)[1]
        for i, pt in enumerate(self.cnt):
            self.mask_labeled_skeleton[pt[1], pt[0]] = labels[i]

    def compute_gradients(self):
        self.gradient_x = cv2.Scharr(255-self.mask, cv2.CV_32F, 1, 0, scale=1)
        self.gradient_y = cv2.Scharr(255-self.mask, cv2.CV_32F, 0, 1, scale=1)
        self.gradient_x = self.gradient_x + 0.0001  # Avoid div by zero

    def compute_angle(self, x1, y1, x2, y2):
        # Angle between gradient vectors (gx1, gy1) and (gx2, gy2)
        gx1 = self.gradient_x[y1, x1]
        gy1 = self.gradient_y[y1, x1]
        gx2 = self.gradient_x[y2, x2]
        gy2 = self.gradient_y[y2, x2]
        cos_angle = gx1 * gx2 + gy1 * gy2
        cos_angle /= (np.linalg.norm((gx1, gy1)) * np.linalg.norm((gx2, gy2)))
        angle = np.arccos(cos_angle)
        if cos_angle < 0:
            angle = np.pi - angle
        return angle

    def draw_curvature_heatmap(self, n=5):
        self.curvature_heatmap = np.zeros_like(self.mask, dtype=np.float)
        contour = self.cnt.squeeze()
        N = len(contour)
        for i in range(N):
            x1, y1 = contour[i]
            x2, y2 = contour[(i + n) % N]
            angle = self.compute_angle(x1, y1, x2, y2)
            # Get the middle point between i and (i + n)
            x3, y3 = contour[((2*i + n) // 2) % N]
            # Use angle between gradient vectors as score
            self.curvature_heatmap[y3, x3] = angle

    # find body end point on contour. n is length of label span of end segment,
    def find_end_points(self, n=10):
        heatmap_end_1 = (self.curvature_heatmap
                         * [self.mask_labeled_skeleton < n+1]
                         ).squeeze()
        heatmap_end_2 = (self.curvature_heatmap
                         * [self.mask_labeled_skeleton
                            > len(self.skeleton)-n-1]
                         ).squeeze()
        self.end_pt_1 = np.unravel_index(
            np.argmax(heatmap_end_1), heatmap_end_1.shape)[::-1]  # (x,y)
        self.end_pt_2 = np.unravel_index(
            np.argmax(heatmap_end_2), heatmap_end_2.shape)[::-1]  # (x,y)

    def make_contour_graph(self):
        self.contour_graph = nx.Graph()
        self.contour_graph.add_nodes_from(
            ['%d,%d' % (pt[0], pt[1]) for pt in self.cnt])  # (x,y)
        for i, pt in enumerate(self.cnt):
            pt_1 = '%d,%d' % (self.cnt[i][0], self.cnt[i][1])  # (x,y)
            pt_2 = '%d,%d' % (self.cnt[i-1][0], self.cnt[i-1][1])  # (x,y)
            self.contour_graph.add_edge(pt_1, pt_2)

    def cut_graph_to_2_sides(self):
        break_pt_1 = list(
            self.contour_graph.edges(
                '%d,%d' % (self.end_pt_1)))[0][1]  # (x,y)
        break_pt_2 = list(
            self.contour_graph.edges('%d,%d' %
                                     (self.end_pt_2)))[0][1]  # (x,y)
        self.contour_graph.remove_edge('%d,%d' % (self.end_pt_1), break_pt_1)
        self.contour_graph.remove_edge('%d,%d' % (self.end_pt_2), break_pt_2)

        path1 = nx.all_simple_paths(self.contour_graph, source='%d,%d' % (
            self.end_pt_1), target=break_pt_2)  # this is a generator
        side1 = next(path1)
        self.side1 = [[int(pt.split(',')[0]), int(pt.split(',')[1])]
                      for pt in side1]
        path2 = nx.all_simple_paths(self.contour_graph,
                                    source=break_pt_1,
                                    target='%d,%d' % (self.end_pt_2)
                                    )  # this is a generator
        side2 = next(path2)
        self.side2 = [[int(pt.split(',')[0]), int(pt.split(',')[1])]
                      for pt in side2]

    def map_skeleton_to_2_sides(self):
        tree_side1 = spatial.KDTree(np.array(self.side1))
        tree_side2 = spatial.KDTree(np.array(self.side2))
        self.labels_side1 = tree_side1.query(np.squeeze(self.skeleton))[1]
        self.labels_side2 = tree_side2.query(np.squeeze(self.skeleton))[1]

    def compute_widths(self):
        self.widths = []
        self.side1_widths = []
        self.side2_widths = []
        for i in range(len(self.skeleton)):
            pt1 = np.array(self.side1[self.labels_side1[i]])
            pt2 = np.array(self.side2[self.labels_side2[i]])
            skeleton_pt = self.skeleton[i]
            self.widths.append(np.linalg.norm(pt1-pt2))
            self.side1_widths.append(np.linalg.norm(pt1-skeleton_pt))
            self.side2_widths.append(np.linalg.norm(pt2-skeleton_pt))

    def estimate_body_orientation(self):
        integral_length = int(self.integral_span*len(self.skeleton))
        area1 = sum(self.widths[:integral_length])
        area2 = sum(self.widths[-integral_length:])
        self.is_head_ahead = area1 > area2
        return area1 > area2

    def draw_2_sides(self):
        mask = np.zeros_like(self.mask)
        for i, pt in enumerate(self.side1):
            mask[pt[1], pt[0]] = 1  # label should start from 1, 0 is bg
        for i, pt in enumerate(self.side2):
            mask[pt[1], pt[0]] = 2  # label should start from 1, 0 is bg
        plt.imshow(mask)
        plt.colorbar()
        plt.show()


if __name__ == '__main__':
    path = "1.jpg"
    img = cv2.imread(path)
    img = cv2.resize(img, (200, 200))
    mask = cv2.imread('mask_0.png', 0)
    mask = cv2.resize(mask, (200, 200))
    from skeleton_extractor import SkeletonExtractor
    skeleton_extractor = SkeletonExtractor(mask)
    skeleton = skeleton_extractor.skeleton
    pixel_annotator = PixelAnnotator(mask, skeleton)
    pixel_annotator.draw_2_sides()
